# Simple LSTM network implemented with PyTorch
# Data: 1s sampled and low pass filtered data

import torch
import torch.nn as nn
from datetime import datetime
import h5py
import os
import numpy as np
import matplotlib.pyplot as plt
import joblib
from sklearn.preprocessing import RobustScaler, MinMaxScaler
from torch.utils.data import TensorDataset, DataLoader
import torch.optim as optim
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from imv_lstm_networks import IMVFullLSTM
run_start = datetime.now()
# setting the parameters
gpu_num = 0
n_samp = 1 # number of timesteps per feature (currently needs to devide the length of the timeseries)
input_dim = n_samp
output_dim = n_samp
hidden_dim = n_samp
batch_size = 64
layer_dim = 3 # number of layers (lstm)
dropout = 0.5
n_epochs = 10
learning_rate = 1e-3
weight_decay = 0.0
day_len = 24*60*6 # 10 s sampling rate
num_days_train = 50 # number of days to use for training
num_days_test = 3 # number of days to use for testing, in this case this is = num_val_test
use_fake_signal = False
early_stopping = False
amp_signal = False
# setting what device to use
device = torch.device("cuda:%d" % gpu_num if torch.cuda.is_available() else "cpu")

# file locations
dataloc = '/data/wiay/2263373r/Etna/'
savedir = '/data/www.astro/2263373r/Etna_LSTM/Etna_IMVFullLSTM/'
os.mkdir(savedir + 'run_'+str(run_start))
saveloc = savedir + 'run_' + str(run_start)+'/'

#saving the parameters to a txt file
f = open(saveloc + 'params.txt',"w+")
f.write(f"Number of layers: {layer_dim} \n")
f.write(f"Dropout: {dropout} \n")
f.write(f"Weight decay: {weight_decay} \n")
f.write(f"Learning rate: {learning_rate} \n")
f.write(f"Number of epochs: {n_epochs} \n")
f.write(f"Batch size: {batch_size} \n")
f.write(f"Using fake signal: {use_fake_signal} \n")
f.write(f"Amplifying signal: {amp_signal} \n")
f.write(f"Early stopping: {early_stopping} \n")
# -------------------------------------FUNCTIONS--------------------------------------------------
class LSTMModel(nn.Module):
    def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
        super().__init__()

        # Defining the number of layers and the nodes in each layer
        self.hidden_dim = hidden_dim
        self.layer_dim = layer_dim

        # LSTM layers
        self.lstm = nn.LSTM(
            input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
        )

        # Fully connected layer
        self.fc = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        # Initializing hidden state for first input with zeros
        h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_().to(device)

        # Initializing cell state for first input with zeros
        c0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_().to(device)

        # We need to detach as we are doing truncated backpropagation through time (BPTT)
        # If we don't, we'll backprop all the way to the start even after going through another batch
        # Forward propagation by passing in the input, hidden state, and cell state into the model
        out, (hn, cn) = self.lstm(x, (h0.detach(), c0.detach()))

        # Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
        # so that it can fit into the fully connected layer
        out = out[:, -1, :]

        # Convert the final state to our desired output shape (batch_size, output_dim)
        out = self.fc(out).to(device)

        return out

class Optimization:
    def __init__(self, model, loss_fn, optimizer):
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.train_losses = []
        self.val_losses = []

    def train_step(self, x, y):
        # Sets model to train mode
        self.model.train()
        # Makes predictions
        yhat = self.model(x)
        print(type(yhat))
        # Computes loss
        loss = self.loss_fn(y, yhat)
        # Computes gradients
        loss.backward()
        # Updates parameters and zeroes gradients
        self.optimizer.step()
        self.optimizer.zero_grad()
        # Returns the loss
        return loss.item()
    def train(self, train_loader, val_loader, test_loader,  batch_size=64, n_epochs=50, n_features=1, saveloc=saveloc,
              run_start=run_start, input_dim=input_dim):
        epochs_no_improve = 0 
        min_val_loss = np.Inf # initially setting the minimum loss to infinity
        n_epochs_stop = 6 # tolerance

        for epoch in range(1, n_epochs + 1):
            batch_losses = []
            for x_batch, y_batch in train_loader:
                x_batch = x_batch.view([batch_size, -1, n_features]).to(device)
                y_batch = y_batch.to(device)
                print(type(x_batch))
                print(type(y_batch))
                loss = self.train_step(x_batch, y_batch)
                batch_losses.append(loss)
            training_loss = np.mean(batch_losses)
            self.train_losses.append(training_loss)

            with torch.no_grad():
                batch_val_losses = []
                for x_val, y_val in val_loader:
                    x_val = x_val.view([batch_size, -1, n_features]).to(device)
                    y_val = y_val.to(device)
                    self.model.eval()
                    yhat = self.model(x_val)
                    val_loss = self.loss_fn(y_val, yhat).item()
                    batch_val_losses.append(val_loss)
                validation_loss = np.mean(batch_val_losses)
                self.val_losses.append(validation_loss)

            # printing losses
            if (epoch <= 10) | (epoch % 10 == 0):
                print(
                    f"[{epoch}/{n_epochs}] Training loss: {training_loss:.4f}\t "
                    f"Validation loss: {validation_loss:.4f}"
                )

            if (epoch <= 10) | (epoch % 10 == 0): # every 50 epochs save the loss plot and model
                self.plot_losses()
                torch.save(self.model.state_dict(), (saveloc + 'model.h5'))
            if (epoch % 100 == 0): # every 100 epochs make a test plot
                predictions, values = self.evaluate(test_loader, n_features=input_dim)
                metrics = self.calculate_metrics(predictions, values)
                self.plot_test(predictions, values, metrics)
            # early stopping condition
            if early_stopping:
                if validation_loss < min_val_loss:
                    epochs_no_improve = 0
                    min_val_loss = validation_loss
                else:
                    epochs_no_improve += 1

                if epoch > 50 and epochs_no_improve == n_epochs_stop:
                    print('Early stopping!')
                    break

    def evaluate(self, test_loader, batch_size=1, n_features=1):
        with torch.no_grad():
            predictions = []
            values = []
            for x_test, y_test in test_loader:
                x_test = x_test.view([batch_size, -1, n_features]).to(device)
                y_test = y_test.to(device)
                self.model.eval()
                yhat = self.model(x_test)
                predictions.append(yhat.cpu().detach().numpy())
                values.append(y_test.cpu().detach().numpy())
        return predictions, values

    def plot_test(self, predictions, values, metrics, run_start=run_start):
        p = np.array(predictions)
        v = np.array(values)
        p = np.reshape(p, (np.shape(p)[0] * np.shape(p)[2]))
        v = np.reshape(v, (np.shape(v)[0] * np.shape(v)[2]))
        t = np.arange(0, len(p)*10, 10)
        plt.plot(t, p, alpha=0.8, label='prediction')
        plt.plot(t, v, label='truth')
        plt.xlabel('time (s)')
        plt.title('Predictions vs Target')
        plt.text(0, -1, f"MAE: {metrics['mae']} \n MSE: {metrics['mse']} \n R2: {metrics['r2']}") 
        plt.legend()
        plt.savefig(saveloc + 'testpred.png')
        plt.close()

    def calculate_metrics(self, predictions, values):
        p = np.array(predictions)
        v = np.array(values)
        p = np.reshape(p, (np.shape(p)[0] * np.shape(p)[2]))
        v = np.reshape(v, (np.shape(v)[0] * np.shape(v)[2]))
        return {'mae': mean_absolute_error(v, p),
                'mse': mean_squared_error(v, p),
                'r2': r2_score(v, p)}

    def plot_losses(self, run_start=run_start, saveloc=saveloc):
        plt.plot(self.train_losses, label="Training loss")
        plt.plot(self.val_losses, label="Validation loss")
        plt.legend()
        plt.ylabel("MSE")
        plt.yscale("log")
        plt.xlabel("Epoch")
        plt.title("Losses")
        plt.savefig(saveloc + 'loss.png')
        plt.close()

def get_model(model, model_params):
    models = {
        "lstm": LSTMModel,
        "imv_full_lstm": IMVFullLSTM
    }
    return models.get(model.lower())(**model_params)

def fake_data(x_data):
    t = np.arange(len(x_data[:, 0]))
    s_1 = np.cos(t/10000)
    s_2 = np.sin(t/20000)
    s_3 = np.cos(t/100000)
    s_4 = np.sin(t/400000)
    y = s_1 + s_2 + s_3 + s_4
    y_scaled = y*np.std(x_data[:, 0]/10)
    new_x = x_data[:, 0] + y_scaled
    # scaling the signal
    scaler = RobustScaler()
    y = scaler.fit_transform(y_scaled.reshape(-1, 1))
    y = np.squeeze(y, axis=1)
    joblib.dump(scaler, "igrav_scaler.z")
    new_x = scaler.fit_transform(new_x.reshape(-1, 1))
    new_x = np.squeeze(new_x, axis=1)
    x_data[:, 0] = new_x
    return y, x_data

def plot_inputs(x_data,y_data,saveloc=saveloc, run_start=run_start):
    t = np.arange(0, np.shape(x_data)[0] * 10, 10)
    plt.figure()
    plt.subplot(np.shape(x_data)[1] + 1, 1, 1)
    plt.plot(t, y_data)
    for i in range(np.shape(x_data)[1]):
        plt.subplot(np.shape(x_data)[1] + 1, 1, i + 2)
        plt.plot(t, x_data[:, i])
    plt.savefig(saveloc+'input_data.png')
    plt.xlabel('time (s)')
    plt.show()
    plt.close()

#---------------------------------LOADING FILES---------------------------------------
n = h5py.File(dataloc + 'dataset_1410_0912.hdf5', 'r')
# channels to use
labels = ['CH4R', 'AD7195_2_Ch1', 'T_DAC1', 'T_DAC4', 'tilt_X']
f.write("Channels used: %s \n" %labels)
# labels = ['CH4R']
x_data = []
# rescaling the data
for label in labels:
    scaler = RobustScaler()
    d = np.array(n[label])
    d = scaler.fit_transform(d.reshape(-1, 1))
    d = np.squeeze(d, axis=1)
    x_data = np.append(x_data, d)
x_data = x_data.reshape(len(labels), int(len(x_data)/len(labels))).T
scaler = RobustScaler()
d = np.array(n['y_data'])
d = scaler.fit_transform(d.reshape(-1, 1))
d = np.squeeze(d, axis=1)
y_data = d
joblib.dump(scaler, "igrav_scaler.z")

# overwriting with fake signal
if use_fake_signal:
    y_data, x_data = fake_data(x_data)

# add the signal into the main channel
if amp_signal:
    new_x = np.squeeze(x_data[:, 0])+y_data/100
    new_x = scaler.fit_transform(new_x.reshape(-1, 1))
    x_data[:, 0] = np.squeeze(new_x)

# plotting in the input data
plot_inputs(x_data,y_data)

# reshape into (samples, timestep_per_sample, features), features = channels
data_X = np.reshape(x_data, (int(np.shape(x_data)[0]/n_samp), n_samp, int(np.shape(x_data)[1])))
data_Y = np.reshape(y_data, (int(np.shape(y_data)[0]/n_samp), n_samp))

#splitting it into train, validation, and test data
len_train=int((day_len*num_days_train)/n_samp)
len_test=int(day_len*num_days_test/n_samp)
train_X, val_X, test_X = data_X[(len_test*2):, :, :], data_X[:len_test, :, :], \
                         data_X[len_test:(len_test*2), :, :]
train_Y, val_Y, test_Y = data_Y[(len_test*2):], data_Y[:len_test], data_Y[len_test:(len_test*2)]
print("Training data shape: x = "+str(np.shape(train_X))+" y = "+str(np.shape(train_Y)))
print("Validation data shape: x = "+str(np.shape(val_X))+" y = "+str(np.shape(val_Y)))
print("Test data shape: x = "+str(np.shape(test_X))+" y = "+str(np.shape(test_Y)))

f.write("Training data shape: x = "+str(np.shape(train_X))+" y = "+str(np.shape(train_Y))+"\n")
f.write("Validation data shape: x = "+str(np.shape(val_X))+" y = "+str(np.shape(val_Y))+"\n")
f.write("Test data shape: x = "+str(np.shape(test_X))+" y = "+str(np.shape(test_Y))+"\n")

train_features = torch.Tensor(train_X)
train_targets = torch.Tensor(train_Y)
val_features = torch.Tensor(val_X)
val_targets = torch.Tensor(val_Y)
test_features = torch.Tensor(test_X)
test_targets = torch.Tensor(test_Y)

train = TensorDataset(train_features, train_targets)
val = TensorDataset(val_features, val_targets)
test = TensorDataset(test_features, test_targets)

# using pytorch Dataloader to split intp batches
# drop_last=True, of the last bit is not a whole batch then drop it
train_loader = DataLoader(train, batch_size=batch_size, shuffle=False, drop_last=True)
val_loader = DataLoader(val, batch_size=batch_size, shuffle=False, drop_last=True)
test_loader = DataLoader(test, batch_size=batch_size, shuffle=False, drop_last=True)
test_loader_one = DataLoader(test, batch_size=1, shuffle=False, drop_last=True)

# ---------------------------------------------------------------------------------------------------------

model_params = {'input_dim': input_dim,
                # 'hidden_dim' : hidden_dim,
                # 'layer_dim' : layer_dim,
                'output_dim' : output_dim,
                'n_units' : 128}
                # 'dropout_prob' : dropout}
model = get_model('imv_full_lstm', model_params)
loss_fn = nn.MSELoss(reduction="mean")
optimizer = optim.Adam(model.parameters(), lr=learning_rate, weight_decay=weight_decay)
model.to(device)
# training
start_train = datetime.now()
opt = Optimization(model=model, loss_fn=loss_fn, optimizer=optimizer)
opt.train(train_loader, val_loader, test_loader_one, batch_size=batch_size, n_epochs=n_epochs, n_features=input_dim)
opt.plot_losses()
end_train = datetime.now()
print("Training time: "+str(end_train-start_train))
f.write("Training time: "+ str(end_train-start_train)+"\n")
# testing
start_test = datetime.now()
predictions, values = opt.evaluate(test_loader_one, batch_size=1, n_features=input_dim)
metrics = opt.calculate_metrics(predictions, values)
opt.plot_test(predictions, values, metrics)
print("Testing metrics:")
print("MAE: ", metrics['mae'], " MSE: ", metrics['mse'], " R2: ", metrics['r2'])
end_test = datetime.now()
print("Testing time: ",  end_test-start_test)
f.write("Testing metrics:")
f.write(f"MAE: {metrics['mae']}, MSE: {metrics['mse']}, R2: {metrics['r2']} \n")
f.write("Testing time: "+str(end_test-start_test)+ "\n")
f.close()


# -------------------------------------- VARIABLE CONTRIBUTION INTERPRETATION ---------------------------------------------
with torch.no_grad():
    mse_val = 0
    preds = []
    true = []
    alphas = []
    betas = []
    for batch_x, batch_y in test_loader:
        batch_x = batch_x.cuda()
        batch_y = batch_y.cuda()
        output, a, b = model(batch_x)
        output = output.squeeze(1)
        preds.append(output.detach().cpu().numpy())
        true.append(batch_y.detach().cpu().numpy())
        alphas.append(a.detach().cpu().numpy())
        betas.append(b.detach().cpu().numpy())
        mse_val += loss(output, batch_y).item()*batch_x.shape[0]
preds = np.concatenate(preds)
true = np.concatenate(true)
alphas = np.concatenate(alphas)
betas = np.concatenate(betas)
alphas = alphas.mean(axis=0)
betas = betas.mean(axis=0)
alphas = alphas[..., 0]
betas = betas[..., 0]
alphas = alphas.transpose(1, 0)

#plotting the importances at each timestep
fig, ax = plt.subplots(figsize=(20, 20))
im = ax.imshow(alphas)
plt.savefig(saveloc+'var_importance.png')
